import { MaterialLoader, Texture } from 'three';
import { MaterialUtils } from './MaterialUtils';
import { DataTransportPayload, DataTransportPayloadUtils } from './DataTransport';
export class MaterialsTransportPayload extends DataTransportPayload {
    type = 'MaterialsTransportPayload';
    materials = new Map();
    materialsJson = new Map();
    multiMaterialNames = new Map();
    cloneInstructions = [];
}
/**
 * Define a structure that is used to ship materials data between main and workers.
 */
export class MaterialsTransportPayloadUtils {
    static packMaterialsTransportPayload(payload, cloneBuffers) {
        const transferables = [];
        DataTransportPayloadUtils.fillTransferables(payload.buffers.values(), transferables, cloneBuffers);
        payload.materialsJson = MaterialUtils.getMaterialsJSON(payload.materials);
        return {
            payload: payload,
            transferables: transferables
        };
    }
    /**
     * @param {MaterialsTransportDef} transportObject
     * @return {MaterialsTransport}
     */
    static unpackMaterialsTransportPayload(payload, transportObject) {
        for (const [k, v] of transportObject.multiMaterialNames.entries()) {
            payload.multiMaterialNames.set(k, v);
        }
        for (const cloneInstruction of transportObject.cloneInstructions) {
            payload.cloneInstructions.push(cloneInstruction);
        }
        payload.cloneInstructions = transportObject.cloneInstructions;
        const materialLoader = new MaterialLoader();
        for (const [k, v] of transportObject.materialsJson.entries()) {
            payload.materials.set(k, materialLoader.parse(v));
        }
    }
    /**
     * Set an object containing named materials.
     * @param {Map<string, Material>} materials
     */
    static setMaterials(payload, materials) {
        for (const [k, v] of materials.entries()) {
            payload.materials.set(k, v);
        }
    }
    /**
     * Removes all textures and null values from all materials
     */
    static cleanMaterials(payload) {
        const clonedMaterials = new Map();
        for (const material of payload.materials.values()) {
            if (typeof material.clone === 'function') {
                const clonedMaterial = material.clone();
                clonedMaterials.set(clonedMaterial.name, MaterialsTransportPayloadUtils.cleanMaterial(clonedMaterial));
            }
        }
        payload.materials = clonedMaterials;
    }
    static cleanMaterial(material) {
        const objToAlter = material;
        for (const [k, v] of Object.entries(objToAlter)) {
            if ((v instanceof Texture || v === null) && Object.prototype.hasOwnProperty.call(material, k)) {
                objToAlter[k] = undefined;
            }
        }
        return material;
    }
    /**
     * Tell whether a multi-material was defined
     * @return {boolean}
     */
    static hasMultiMaterial(payload) {
        return payload.multiMaterialNames.size > 0;
    }
    /**
     * Returns a single material if it is defined or null.
     * @return {Material|null}
     */
    static getSingleMaterial(payload) {
        return payload.materials.size > 0 ? payload.materials.values().next().value : undefined;
    }
    /**
     * Adds contained material or multi-material the provided materials object or it clones and adds new materials according clone instructions.
     *
     * @param {Map<string, Material>} materials
     * @param {boolean} log
     *
     * @return {Material|Material[]|undefined}
     */
    static processMaterialTransport(payload, materials, log) {
        if (!payload) {
            return undefined;
        }
        for (const cloneInstruction of payload.cloneInstructions) {
            MaterialUtils.cloneMaterial(materials, cloneInstruction, log);
        }
        if (MaterialsTransportPayloadUtils.hasMultiMaterial(payload)) {
            // multi-material
            const outputMaterials = [];
            for (const [k, v] of payload.multiMaterialNames.entries()) {
                const mat = materials.get(v);
                outputMaterials[k] = mat ? mat : undefined;
            }
            return outputMaterials;
        }
        else {
            const singleMaterial = MaterialsTransportPayloadUtils.getSingleMaterial(payload);
            if (singleMaterial) {
                const outputMaterial = materials.get(singleMaterial.name);
                return outputMaterial ? outputMaterial : singleMaterial;
            }
        }
        return undefined;
    }
}
//# sourceMappingURL=MaterialsTransport.js.map