import { Box3, BufferAttribute, BufferGeometry, Sphere } from 'three';
import { DataTransportPayload, DataTransportPayloadUtils } from './DataTransport';
import { MaterialsTransportPayload, MaterialsTransportPayloadUtils } from './MaterialsTransport';
export class MeshTransportPayload extends DataTransportPayload {
    type = 'MeshTransportPayload';
    // 0: mesh, 1: line, 2: point
    geometryType = 0;
    bufferGeometry;
    meshName;
    materialsTransportPayload = new MaterialsTransportPayload();
}
export class MeshTransportPayloadUtils {
    /**
     * Set the {@link BufferGeometry} and geometry type that can be used when a mesh is created.
     *
     * @param {BufferGeometry} bufferGeometry
     * @param {number} geometryType [0=Mesh|1=LineSegments|2=Points]
     */
    static setBufferGeometry(payload, bufferGeometry, geometryType) {
        payload.bufferGeometry = bufferGeometry;
        payload.geometryType = geometryType;
    }
    /**
     * Sets the mesh and the geometry type [0=Mesh|1=LineSegments|2=Points]
     * @param {Mesh} mesh
     * @param {number} geometryType
     */
    static setMesh(payload, mesh, geometryType) {
        payload.meshName = mesh.name;
        MeshTransportPayloadUtils.setBufferGeometry(payload, mesh.geometry, geometryType);
    }
    /**
     * @param {boolean} cloneBuffers
     */
    static packMeshTransportPayload(payload, cloneBuffers) {
        MeshTransportPayloadUtils.packGeometryBuffers(cloneBuffers, payload.bufferGeometry, payload.buffers);
        const transferables = [];
        DataTransportPayloadUtils.fillTransferables(payload.buffers.values(), transferables, cloneBuffers);
        if (payload.materialsTransportPayload) {
            MaterialsTransportPayloadUtils.packMaterialsTransportPayload(payload.materialsTransportPayload, cloneBuffers);
        }
        return {
            payload: payload,
            transferables: transferables
        };
    }
    /**
     * Package {@link BufferGeometry} and prepare it for transport.
     *
     * @param {boolean} cloneBuffers Clone buffers if their content shall stay in the current context.
     */
    static packGeometryBuffers(cloneBuffers, bufferGeometry, buffers) {
        // fast-fail
        if (!(bufferGeometry instanceof BufferGeometry))
            return;
        const vertexBA = bufferGeometry.getAttribute('position');
        const normalBA = bufferGeometry.getAttribute('normal');
        const uvBA = bufferGeometry.getAttribute('uv');
        const colorBA = bufferGeometry.getAttribute('color');
        const skinIndexBA = bufferGeometry.getAttribute('skinIndex');
        const skinWeightBA = bufferGeometry.getAttribute('skinWeight');
        const indexBA = bufferGeometry.getIndex();
        MeshTransportPayloadUtils.addAttributeToBuffers('position', vertexBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('normal', normalBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('uv', uvBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('color', colorBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('skinIndex', skinIndexBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('skinWeight', skinWeightBA, cloneBuffers, buffers);
        MeshTransportPayloadUtils.addAttributeToBuffers('index', indexBA, cloneBuffers, buffers);
    }
    static addAttributeToBuffers(name, input, cloneBuffer, buffers) {
        if (input && input !== null) {
            const arrayLike = (cloneBuffer ? Array.from(input.array).slice(0) : input.array);
            buffers.set(name, arrayLike);
        }
    }
    static unpackMeshTransportPayload(payload, cloneBuffers) {
        const mtp = Object.assign(new MeshTransportPayload(payload.cmd, payload.id), payload);
        mtp.bufferGeometry = MeshTransportPayloadUtils.reconstructBuffer(cloneBuffers, mtp.bufferGeometry);
        if (payload.materialsTransportPayload) {
            mtp.materialsTransportPayload = Object.assign(new MaterialsTransportPayload(), payload.materialsTransportPayload);
            MaterialsTransportPayloadUtils.unpackMaterialsTransportPayload(mtp.materialsTransportPayload, payload.materialsTransportPayload);
        }
        return mtp;
    }
    /**
     * Reconstructs the {@link BufferGeometry} from the raw buffers.
     * @param {boolean} cloneBuffers
     */
    static reconstructBuffer(cloneBuffers, transferredGeometry) {
        const bufferGeometry = new BufferGeometry();
        // fast-fail: transferredGeometry is either rubbish or already a bufferGeometry
        if (!transferredGeometry) {
            return bufferGeometry;
        }
        else if (transferredGeometry instanceof BufferGeometry) {
            return transferredGeometry;
        }
        if (transferredGeometry.attributes) {
            const attr = transferredGeometry.attributes;
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.position, 'position', cloneBuffers);
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.normal, 'normal', cloneBuffers);
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.uv, 'uv', cloneBuffers);
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.color, 'color', cloneBuffers);
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.skinIndex, 'skinIndex', cloneBuffers);
            MeshTransportPayloadUtils.assignAttributeFromTransfered(bufferGeometry, attr.skinWeight, 'skinWeight', cloneBuffers);
        }
        // TODO: morphAttributes
        if (transferredGeometry.index !== null) {
            const indexAttr = transferredGeometry.index;
            if (indexAttr) {
                const indexBuffer = cloneBuffers ? indexAttr.array.slice(0) : indexAttr.array;
                bufferGeometry.setIndex(new BufferAttribute(indexBuffer, indexAttr.itemSize, indexAttr.normalized));
            }
        }
        const boundingBox = transferredGeometry.boundingBox;
        if (boundingBox !== null) {
            bufferGeometry.boundingBox = Object.assign(new Box3(), boundingBox);
        }
        const boundingSphere = transferredGeometry.boundingSphere;
        if (boundingSphere !== null) {
            bufferGeometry.boundingSphere = Object.assign(new Sphere(), boundingSphere);
        }
        bufferGeometry.uuid = transferredGeometry.uuid;
        bufferGeometry.name = transferredGeometry.name;
        bufferGeometry.type = transferredGeometry.type;
        bufferGeometry.groups = transferredGeometry.groups;
        bufferGeometry.drawRange = transferredGeometry.drawRange;
        bufferGeometry.userData = transferredGeometry.userData;
        return bufferGeometry;
    }
    static assignAttributeFromTransfered(bufferGeometry, input, attrName, cloneBuffer) {
        if (input) {
            const arrayLike = cloneBuffer ? input.array.slice(0) : input.array;
            bufferGeometry.setAttribute(attrName, new BufferAttribute(arrayLike, input.itemSize, input.normalized));
        }
    }
}
//# sourceMappingURL=MeshTransport.js.map